@@ -6,6 +6,9 @@ import subprocess
 import os
 import requests
 from . import process_code_template
+from . import downloader
+import json
+import shutil
 
 def get_trtpy_str_variables(prefix="trtpy."):
     trtpy_variables = []
@@ -58,7 +61,6 @@ def do_local_cpp_pkg(args):
 def do_get_cpp_pkg(args):
     name = args.name
     url = f"{trtpy.pypi_base_url}/cpp-packages/{name}.zip"
-    from . import downloader
 
     file = os.path.join(downloader.CACHE_ROOT, "cpp-packages", f"{name}.zip")
     ok, md5 = downloader.download_and_verify_md5_saveto_file(url, file)
@@ -170,8 +172,6 @@ def get_inv_variable_template():
     return variable_template
 
 def do_get_templ(args):
-    from . import downloader
-
     if args.saveto is None:
         args.saveto = args.template
 
@@ -193,14 +193,15 @@ def do_get_templ(args):
 
     trtpy.init(cuda_version=args.cuda, tensorrt_version=args.trt, load_lean_library=False)
     url = f"{trtpy.pypi_base_url}/code_template/{args.template}.zip"
-    to = os.path.join(downloader.CACHE_ROOT, "code_template", f"{args.template}.zip")
-    if not downloader.download_to_file(url, to):
+    cache_zip = os.path.join(downloader.CACHE_ROOT, "code_template", f"{args.template}.zip")
+    if not args.download and os.path.exists(cache_zip):
+        print(f"Use cache {cache_zip}")
+    elif not downloader.download_to_file(url, cache_zip):
         print(f"Template '{args.template}' not found")
         return 1
 
     print(f"Extract to {args.saveto} . ")
-    namelist = downloader.extract_zip_to(to, args.saveto)
-    os.remove(to)
+    namelist = downloader.extract_zip_to(cache_zip, args.saveto)
 
     if not args.raw:
         print("Replace project variable")
@@ -210,6 +211,156 @@ def do_get_templ(args):
     print("Done!")
     return 0
 
+def remove_tree_keep_father(path):
+    files = os.listdir(path)
+    for file in files:
+        full_path = os.path.join(path, file)
+        print(f"Delete {full_path}")
+        if os.path.isdir(full_path):
+            shutil.rmtree(full_path)
+        else:
+            os.remove(full_path)
+
+def extract_series_templ(meta, saveto, raw, download):
+
+    series_name = meta["name"]
+    current_index = meta["current-index"]
+    templ_list = meta["list-templ"]
+    select_item = templ_list[current_index]
+    item_name = select_item[0].strip()
+    if item_name == "" or item_name == "." or item_name == ".." or item_name.find("/") != -1:
+        print(f"Invalid series templ name [{item_name}]")
+        return
+
+    templ_name = f"{series_name}-{item_name}"
+    templ_url = f"{trtpy.pypi_base_url}/code_template/{templ_name}.zip"
+    templ_cache_local_zip = os.path.join(downloader.CACHE_ROOT, "code_template", f"{templ_name}.zip")
+    if not download and os.path.exists(templ_cache_local_zip):
+        print(f"Use cache {templ_cache_local_zip}")
+    elif not downloader.download_to_file(templ_url, templ_cache_local_zip):
+        print(f"Template '{templ_name}' not found")
+        return 1
+
+    if os.path.isdir(saveto):
+        print(f"Remove old files")
+        remove_tree_keep_father(saveto)
+
+    print(f"Extract to {saveto} . ")
+    namelist = downloader.extract_zip_to(templ_cache_local_zip, saveto)
+
+    if not raw:
+        print("Replace project variable")
+        variable_template = get_variable_template()
+        process_code_template.process_code_template(saveto, namelist, variable_template)
+
+    series_directory = os.path.join(saveto, ".series")
+    os.makedirs(series_directory, exist_ok=True)
+    local_config_json = os.path.join(series_directory, "config.json")
+
+    print(f"Save to {local_config_json}")
+    json.dump(meta, open(local_config_json, "w"), indent=4, ensure_ascii=False)
+    print(f"Done! current template is {templ_name}, {select_item[1]}")
+    return 0
+
+def do_get_series(args):
+    if args.saveto is None:
+        args.saveto = args.name
+
+    if os.path.isdir(args.saveto):
+        while True:
+            opt = input(f"{args.saveto} has exists, overwrite? (Y=yes, N=no): default [Y]:").lower()
+            if opt == "": opt = "y"
+            if opt != "y" and opt != "n":
+                continue
+                
+            if opt == "n":
+                print("Operation cancel.")
+                return 0
+            break
+
+    if os.path.isfile(args.saveto):
+        print(f"{args.saveto} is file")
+        return 1
+
+    trtpy.init(cuda_version=args.cuda, tensorrt_version=args.trt, load_lean_library=False)
+    meta_url = f"{trtpy.pypi_base_url}/code_template/{args.name}.series.json"
+    meta_cache_local_file = os.path.join(downloader.CACHE_ROOT, "code_template", f"{args.name}.series.json")
+    if not args.download and os.path.exists(meta_cache_local_file):
+        print(f"Use cache {meta_cache_local_file}")
+    elif not downloader.download_to_file(meta_url, meta_cache_local_file):
+        print(f"Series '{args.name}' not found")
+        return 1
+
+    meta = json.load(open(meta_cache_local_file))
+    templ_list = meta["list-templ"]
+
+    current_index = 0
+    meta["current-index"] = current_index
+    if len(templ_list) == 0:
+        print(f"Series list-templ is empty")
+        return 0
+
+    return extract_series_templ(meta, args.saveto, args.raw, args.download)
+
+
+def do_change_proj(args):
+
+    current_dir = os.path.abspath(".")
+    config_json_file = os.path.join(current_dir, ".series", "config.json")
+    if not os.path.isfile(config_json_file):
+        print(f"Current is not series folder")
+        return 1
+
+    trtpy.init(cuda_version=args.cuda, tensorrt_version=args.trt, load_lean_library=False)
+    meta = json.load(open(config_json_file, "r"))
+    current_index = meta["current-index"]
+    templ_list = meta["list-templ"]
+    if len(templ_list) == 0:
+        print(f"list templ is empty")
+        return 1
+
+    if args.index in ["next", "prev"]:
+        if args.index == "next":
+            current_index = max(0, current_index)
+            current_index = (current_index + 1) % len(templ_list)
+        elif args.index == "prev":
+            current_index = max(0, current_index)
+            current_index = (current_index - 1 + len(templ_list)) % len(templ_list)
+        
+        meta["current-index"] = current_index
+        return extract_series_templ(meta, current_dir, args.raw, args.download)
+
+    index_names_map = []
+    for index, (templ_name, templ_descript) in enumerate(templ_list):
+        p = templ_name.find("-")
+        if p != -1:
+            names = [templ_name[:p], templ_name[p+1:]]
+        else:
+            names = [templ_name]    
+        
+        names = [name for name in names if name.strip() != ""]
+        index_names_map.append([index, names, templ_descript])
+
+    if args.index != "":
+        for index, names, templ_descript in index_names_map:
+            if args.index in names:
+                meta["current-index"] = index
+                return extract_series_templ(meta, current_dir, args.raw, args.download)
+    
+    print(f"List templ: ")
+    for index, names, templ_descript in index_names_map:
+        if len(names) > 0:
+            if len(names) == 1:
+                print(names[0])
+            else:
+                chapter = names[0]
+                caption = names[1]
+                print(f"chapter: {chapter}, caption: {caption}, description: {templ_descript}")
+        
+    if args.index != "":
+        print(f"Unknow index [{args.index}]")
+    return 1
+
 
 def do_templ_list(args):
 
@@ -392,52 +543,51 @@ def do_invert_variable(args):
 
 if __name__ == "__main__":
     do_if_exec()
+
+    def new_parser_with_version(name, help):
+        global subp
+        p = subp.add_parser(name, help=help)
+        p.add_argument("--cuda", type=str, help="cuda version, 10.2, 11.2, 10, 11 etc.", default=None)
+        p.add_argument("--trt", type=str, help="trt version, 8.0, 8 etc.", default=None)
+        return p
+
     parser = argparse.ArgumentParser()
     subp = parser.add_subparsers(dest="cmd")
-    downloadp = subp.add_parser("get-env", help="download environment")
-    downloadp.add_argument("--cuda", type=str, help="cuda version, 10.2, 11.2, 10, 11 etc.", default=None)
-    downloadp.add_argument("--trt", type=str, help="trt version, 8.0, 8 etc.", default=None)
+    new_parser_with_version("get-env", help="download environment")
+    new_parser_with_version("info", help="display support list")
+    new_parser_with_version("mnist-test", help="test tensorrt with mnist")
     
-    execp = subp.add_parser("exec", help="same to ./trtexec")
-    execp.add_argument("--cuda", type=str, help="cuda version, 10.2, 11.2, 10, 11 etc.", default=None)
-    execp.add_argument("--trt", type=str, help="trt version, 8.0, 8 etc.", default=None)
-    execp.add_argument("args", type=str, help="./trtexec args", default=None)
+    p = new_parser_with_version("exec", help="same to ./trtexec")
+    p.add_argument("args", type=str, help="./trtexec args", default=None)
 
     subp.add_parser("list-templ", help="display all code template")
-    templ_search_p = subp.add_parser("search-templ", help="search all code template")
-    templ_search_p.add_argument("pattern", type=str, help="search name, yolo* etc.")
+    p = subp.add_parser("search-templ", help="search all code template")
+    p.add_argument("pattern", type=str, help="search name, yolo* etc.")
 
-    p = subp.add_parser("get-templ", help="fetch code template")
+    p = new_parser_with_version("get-templ", help="fetch code template")
     p.add_argument("template", type=str, help="template name: tensorrt-mnist cuda-sample")
     p.add_argument("saveto", type=str, help="save to directory, default[template name]", nargs="?")
     p.add_argument("--raw", action="store_true", help="do not replace variables")
-    p.add_argument("--cuda", type=str, help="cuda version, 10.2, 11.2, 10, 11 etc.", default=None)
-    p.add_argument("--trt", type=str, help="trt version, 8.0, 8 etc.", default=None)
+    p.add_argument("--download", action="store_true", help="ignore cache and download template")
 
-    cpp_envp = subp.add_parser("get-cpp-pkg", help="download cpp package")
-    cpp_envp.add_argument("name", type=str, help="package name")
+    p = new_parser_with_version("get-series", help="fetch series template")
+    p.add_argument("name", type=str, help="series name")
+    p.add_argument("saveto", type=str, help="save to directory, default[template name]", nargs="?")
+    p.add_argument("--raw", action="store_true", help="do not replace variables")
+    p.add_argument("--download", action="store_true", help="ignore cache and download template")
 
-    cpp_env_listp = subp.add_parser("local-cpp-pkg", help="display all installed local cpp package")
+    p = new_parser_with_version("change-proj", help="change series proj")
+    p.add_argument("index", type=str, default="", help="series proj index, 1.1/cuinit", nargs="?")
+    p.add_argument("--raw", action="store_true", help="do not replace variables")
+    p.add_argument("--download", action="store_true", help="ignore cache and download template")
 
-    p = subp.add_parser("info", help="display support list")
-    p.add_argument("--cuda", type=str, help="cuda version, 10.2, 11.2, 10, 11 etc.", default=None)
-    p.add_argument("--trt", type=str, help="trt version, 8.0, 8 etc.", default=None)
-
-    p = subp.add_parser("prep-vars", help="replace local file variables")
+    p = new_parser_with_version("prep-vars", help="replace local file variables")
     p.add_argument("file_or_directory", type=str, help=f"Project directory or file, file filter = {process_code_template.include_list}")
-    p.add_argument("--cuda", type=str, help="cuda version, 10.2, 11.2, 10, 11 etc.", default=None)
-    p.add_argument("--trt", type=str, help="trt version, 8.0, 8 etc.", default=None)
 
-    p = subp.add_parser("inv-vars", help="replace local file value invert to variables")
+    p = new_parser_with_version("inv-vars", help="replace local file value invert to variables")
     p.add_argument("file_or_directory", type=str, help=f"Project directory or file, file filter = {process_code_template.include_list}")
-    p.add_argument("--cuda", type=str, help="cuda version, 10.2, 11.2, 10, 11 etc.", default=None)
-    p.add_argument("--trt", type=str, help="trt version, 8.0, 8 etc.", default=None)
 
-    p = subp.add_parser("mnist-test", help="test tensorrt with mnist")
-    p.add_argument("--cuda", type=str, help="cuda version, 10.2, 11.2, 10, 11 etc.", default=None)
-    p.add_argument("--trt", type=str, help="trt version, 8.0, 8 etc.", default=None)
-
-    p = subp.add_parser("compile", help="compile use trtpro")
+    p = new_parser_with_version("compile", help="compile use trtpro")
     p.add_argument("onnx", type=str, help="onnx file")
     p.add_argument("--out", type=str, default=".", help="output file or directory")
     p.add_argument("--fp16", action="store_true", help="use fp16")
@@ -445,20 +595,18 @@ if __name__ == "__main__":
     p.add_argument("--onnx-from-hub", action="store_true", default=0, help="use onnx hub file")
     p.add_argument("--max-batch-size", type=int, default=1, help="max batch size")
     p.add_argument("--max-workspace-size", type=int, default=1<<30, help="max workspace size")
-    p.add_argument("--cuda", type=str, help="cuda version, 10.2, 11.2, 10, 11 etc.", default=None)
-    p.add_argument("--trt", type=str, help="trt version, 8.0, 8 etc.", default=None)
 
-    p = subp.add_parser("tryload", help="try to load trtmodel file")
+    p = new_parser_with_version("tryload", help="try to load trtmodel file")
     p.add_argument("trtmodel", type=str, help="tensorRT engine file")
     p.add_argument("--device", type=int, default=0, help="device id")
-    p.add_argument("--cuda", type=str, help="cuda version, 10.2, 11.2, 10, 11 etc.", default=None)
-    p.add_argument("--trt", type=str, help="trt version, 8.0, 8 etc.", default=None)
 
-    p = subp.add_parser("env-source", help="get trtpy environment")
+    p = new_parser_with_version("env-source", help="get trtpy environment")
     p.add_argument("--save", type=str, default="trtpy-environment.sh", help="save to file")
     p.add_argument("--print", action="store_true", help="print env path")
-    p.add_argument("--cuda", type=str, help="cuda version, 10.2, 11.2, 10, 11 etc.", default=None)
-    p.add_argument("--trt", type=str, help="trt version, 8.0, 8 etc.", default=None)
+
+    subp.add_parser("local-cpp-pkg", help="display all installed local cpp package")
+    p = subp.add_parser("get-cpp-pkg", help="download cpp package")
+    p.add_argument("name", type=str, help="package name")
     args = parser.parse_args()
 
     cmd_funcs = {
@@ -474,7 +622,9 @@ if __name__ == "__main__":
         "compile":       do_compile,
         "tryload":       do_tryload,
         "env-source":    do_env_source,
-        "inv-vars":      do_invert_variable
+        "inv-vars":      do_invert_variable,
+        "get-series":    do_get_series,
+        "change-proj":   do_change_proj
     }
 
     if args.cmd in cmd_funcs:
